import ctypes
import sys
import time
from os.path import join as pjoin

import cv2
import numpy as np

from common import constants
from python_wrapper import kdp_wrapper
import kdp_host_api as api


KDP_UART_DEV = 0
KDP_USB_DEV = 1

FW_SIZE = 512
MODEL_SIZE = 20 * 1024 * 1024

LOOPS = 1

IMG_SOURCE_W = 128 
IMG_SOURCE_H = 128
DME_IMG_SIZE = IMG_SOURCE_W * IMG_SOURCE_H * 2


def load_model(dev_idx, model_dir, model_id=1000):
    ret_size = 0
    data = (ctypes.c_char * FW_SIZE)()
    p_buf = (ctypes.c_char * MODEL_SIZE)()

    print("loading models to Kneron Device: ")
    n_len = api.read_file_to_buf(
        data, pjoin(model_dir, 'fw_info.bin'), FW_SIZE)
    if n_len <= 0:
        print("reading fw setup file failed: {}...\n".format(n_len))
        return -1

    dat_size = n_len
    print(dat_size)

    n_len = api.read_file_to_buf(
        p_buf, pjoin(model_dir, 'all_models.bin'), MODEL_SIZE)
    if n_len <= 0:
        print("reading model file failed: {}...\n".format(n_len))
        return -1

    buf_len = n_len
    model_size = n_len
    print(buf_len, model_size)

    print("starting DME mode ...\n")
    ret, ret_size = api.kdp_start_dme(
        dev_idx, model_size, data, dat_size, ret_size, p_buf, buf_len)
    if ret:
        print("could not set to DME mode:{}..\n".format(ret_size))
        return -1
    time.sleep(0.1)

    # dme configuration
    output_num = 1     # number of output node for the model
    image_col = 128
    image_row = 128
    image_ch = 3
    image_format = (constants.IMAGE_FORMAT_SUB128 |
                    constants.NPU_FORMAT_RGB565 |
                    constants.IMAGE_FORMAT_RAW_OUTPUT |
                    constants.IMAGE_FORMAT_CHANGE_ASPECT_RATIO)

    c_int_p = ctypes.POINTER(ctypes.c_int)
    crop_box = np.array([0, 0, 0, 0])    # for future use
    crop_box = crop_box.astype(np.int32)
    crop_box = crop_box.ctypes.data_as(c_int_p)
    pad_value = np.array([0, 0, 0, 0])   # for future use
    pad_value = pad_value.astype(np.int32)
    pad_value = pad_value.ctypes.data_as(c_int_p)
    c_float_p = ctypes.POINTER(ctypes.c_float)
    extra_param = np.array([0.0])
    extra_param = extra_param.astype(np.float32)
    extra_param = extra_param.ctypes.data_as(c_float_p)
    dme_cfg = constants.KDPDMEConfig(
        model_id, output_num, image_col, image_row, image_ch, image_format,
        crop_box, pad_value, extra_param)
    print(f'DME Config: {dme_cfg}')

    dat_size = ctypes.sizeof(dme_cfg)
    print("starting DME configure ...\n")
    ret, model_id = api.kdp_dme_configure(
        dev_idx,
        ctypes.cast(ctypes.byref(dme_cfg), ctypes.c_char_p),
        dat_size, model_id)
    if ret:
        print("could not set to DME configure mode..\n")
        return -1
    time.sleep(0.1)


def read_image(img_path):
    img = cv2.imread(img_path)
    img = cv2.resize(
        img, (IMG_SOURCE_W, IMG_SOURCE_H), interpolation=cv2.INTER_CUBIC)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2BGR565)
    img_data = img.reshape(DME_IMG_SIZE)
    buf_len = DME_IMG_SIZE
    c_char_p = ctypes.POINTER(ctypes.c_char)
    img_data = img_data.astype(np.uint8)
    data_p = img_data.ctypes.data_as(c_char_p)

    return data_p, buf_len


if __name__ == '__main__':
    api.kdp_init_log("/tmp/", "dt42.log")

    if api.kdp_lib_init() < 0:
        print("init for kdp host lib failed.")
    print("adding devices...")
    dev_idx = api.kdp_add_dev(KDP_USB_DEV, "")
    print(f'dev id: {dev_idx}')
    if dev_idx < 0:
        print("add device failed.")
    print("start kdp host lib...")
    if api.kdp_lib_start() < 0:
        print("start kdp host lib failed.")

    model_dir = sys.argv[1]
    img_path = sys.argv[2]

    model_id = 1000
    load_model(dev_idx, model_dir, model_id)
    img_buf, buf_len = read_image(img_path)

    inf_size = 0
    inf_res = (ctypes.c_char * 256000)()
    res_flag = False
    mode = 0
    status = 0
    time_list = []
    for i in range(LOOPS):
        beg = time.time()
        _ret, ssid, res_flag = api.kdp_dme_inference(dev_idx,
                                                     img_buf,
                                                     buf_len,
                                                     inf_size,
                                                     res_flag,
                                                     inf_res,
                                                     mode,
                                                     model_id)
        time_list.append(time.time() - beg)

    avg_time = sum(time_list) / LOOPS
    print(f'Avg time: {avg_time}, frame_time: {1.0 / avg_time}')

    api.kdp_dme_retrieve_res(dev_idx, 0, inf_size, inf_res)
    print(inf_res[:128])

    fp_header_res = ctypes.cast(
        ctypes.byref(inf_res),
        ctypes.POINTER(constants.RawFixpointData)).contents
    output_num = fp_header_res.output_num
    print(output_num, fp_header_res.out_node_params[:])
    # fp_header_res = ctypes.cast(
    #     ctypes.byref(inf_res),
    #     ctypes.POINTER(ctypes.c_float)).contents
    # print(fp_header_res)

    kdp_wrapper.kdp_exit_dme(dev_idx)
    print("deinit kdp host lib...")
    api.kdp_lib_de_init()
